@@ -14,6 +14,7 @@ import logging
 import sys
 import threading
 import traceback
+from collections import defaultdict
 from dataclasses import dataclass, field
 from inspect import isclass
 from types import TracebackType
@@ -228,6 +229,21 @@ class Element(Generic[W]):
             assert frame_py is not None
             self.traceback = TracebackType(tb_frame=frame_py, tb_lasti=self.traceback.tb_lasti, tb_lineno=frame_py.f_lineno, tb_next=None)
 
+    def _arguments_changed(self, other: "Element"):
+        if len(self.args) != len(other.args):
+            return True
+        if len(self.kwargs) != len(other.kwargs):
+            return True
+        for k, v in self.kwargs.items():
+            if k not in other.kwargs:
+                return True
+            if not utils.equals(v, other.kwargs[k]):
+                return True
+        for a, b in zip(self.args, other.args):
+            if not utils.equals(a, b):
+                return True
+        return False
+
     def key(self, value: str):
         """Returns the same element with a custom key set.
 
@@ -390,7 +406,7 @@ class Element(Generic[W]):
         # it's an event listener
         if callback != callback_prev and callback_prev is not None:
             self._remove_widget_event_listener(widget, name, callback_prev)
-        if callback is not None:
+        if callback is not None and callback != callback_prev:
             self._add_widget_event_listener(widget, name, callback)
 
     def _add_widget_event_listener(self, widget: widgets.Widget, name: str, callback: Callable):
@@ -529,6 +545,7 @@ class ComponentFunction(Component):
         self.name = self.f.__name__
         self.mime_bundle = mime_bundle
         self.value_name = value_name
+        self.render_count = 0  # used for debugging, not thread safe, so count can be off
         functools.update_wrapper(self, f)
 
     def widget(self, **kwargs):
@@ -842,7 +859,11 @@ class UserContext(Generic[T]):
         rc = _get_render_context()
         context = rc.context
         assert context is not None
+        prev = context.user_contexts_prev.get(self, self._default_value)
         context.user_contexts[self] = obj
+        if not utils.equals(prev, obj):
+            for listener in context.context_listeners.get(self, []):
+                listener()
 
     def __repr__(self):
         return f"UserContext({self._default_value}, name={self.name})"
@@ -858,18 +879,65 @@ def provide_context(user_context: UserContext[T], obj: T):
     user_context.provide(obj)
 
 
-def use_context(user_context: UserContext[T]) -> T:
+def get_context(user_context: UserContext[T]) -> T:
+    """Similar to use_context, but does not trigger a re-render if the context changes."""
     rc = _get_render_context()
     value = None
-    context = rc.context
+    assert rc.context is not None
+    context = rc.context.parent
+    # we need to walk up the context tree to find the nearest
+    # ancestor that has the context we are looking for.
     while value is None and context is not None:
         if user_context in context.user_contexts:
             value = context.user_contexts.get(user_context)
-            return cast(T, value)
+            break
         else:
             context = context.parent
 
-    return user_context._default_value
+    if context is None:
+        # we did not find the context, so we return the default value
+        return user_context._default_value
+    else:
+        return cast(T, value)
+
+
+def use_context(user_context: UserContext[T]) -> T:
+    counter, set_counter = use_state(0)
+
+    def force_update():
+        set_counter(lambda x: x + 1)
+
+    rc = _get_render_context()
+    value = None
+    assert rc.context is not None
+    context = rc.context.parent
+    # we need to walk up the context tree to find the nearest
+    # ancestor that has the context we are looking for.
+    while value is None and context is not None:
+        if user_context in context.user_contexts:
+            value = context.user_contexts.get(user_context)
+            break
+        else:
+            context = context.parent
+
+    # we need to register a listener on the context so that we can
+    # force an update when the context value changes (using provides).
+    def connect():
+        if context is not None:
+            context.context_listeners[user_context].add(force_update)
+
+            def cleanup():
+                assert context is not None
+                context.context_listeners[user_context].remove(force_update)
+
+            return cleanup
+
+    use_effect(connect, dependencies=[context])
+    if context is None:
+        # we did not find the context, so we return the default value
+        return user_context._default_value
+    else:
+        return cast(T, value)
 
 
 """
@@ -929,11 +997,14 @@ class ComponentContext:
     memo: List[Any] = field(default_factory=list)
     memo_index = 0
     # for provide/use_context
-    user_contexts: Dict[Any, Any] = field(default_factory=dict)
+    user_contexts: Dict["UserContext", Any] = field(default_factory=dict)
+    user_contexts_prev: Dict["UserContext", Any] = field(default_factory=dict)
+    context_listeners: Dict["UserContext", Set[Callable]] = field(default_factory=lambda: defaultdict(set))
 
     # to track key collisions, and remove unused elements
     used_keys: Set[str] = field(default_factory=set)
-    # needs_render: bool = False
+    # if a child component's state if changed, it needs a rerender
+    needs_render: bool = True
 
     # elements created in this context go there
     owns: Set[Element] = field(default_factory=set)
@@ -954,6 +1025,15 @@ class ComponentContext:
 TEffect = TypeVar("TEffect", bound="Effect")
 
 
+@dataclass
+class RerenderReason:
+    reason: str
+    prev_value: Any = None
+    next_value: Any = None
+    created_stack: List[str] = field(default_factory=list)
+    trigger_stack: List[str] = field(default_factory=list)
+
+
 class Effect:
     def __init__(self, callable: EffectCallable, dependencies: Optional[List[Any]] = None, next: Optional["Effect"] = None) -> None:
         self.callable = callable
@@ -993,7 +1073,7 @@ class _RenderContext:
         self.last_root_widget: widgets.Widget = None
         self._is_rendering = False
         self._rerender_needed = False
-        self._rerender_needed_reason: Optional[str] = None
+        self._rerender_needed_reasons: List[RerenderReason] = []
         self.thread_lock = threading.Lock()
         self._closing = False
         self.tracebacks: List[TracebackType] = []
@@ -1043,6 +1123,9 @@ class _RenderContext:
         if self._orphans:
             orphan_widgets = set([widgets.Widget.widgets[k] for k in self._orphans])
             raise RuntimeError(f"Orphan widgets not cleaned up for widgets: {orphan_widgets}")
+        exceptions = [*self.context.exceptions_children, *self.context_root.exceptions_self]
+        if exceptions:
+            raise exceptions[0]
 
     def state_get(self, context: Optional[ComponentContext] = None):
         if context is None:
@@ -1103,6 +1186,9 @@ class _RenderContext:
             return state, self.make_setter(key, self.context, eq)
 
     def make_setter(self, key, context: ComponentContext, eq: Callable[[Any, Any], bool] = None):
+        if DEBUG:
+            created_stack = traceback.format_stack()
+
         def set_(value):
             if callable(value):
                 value = value(context.state[key])
@@ -1142,6 +1228,7 @@ class _RenderContext:
                 should_update = False
 
             if should_update:
+                prev_value = context.state[key]
                 context.state[key] = value
                 if context.state[key] is value and isinstance(value, (list, dict, set)) and new_metadata is None:
                     new_metadata = len(value)
@@ -1149,10 +1236,23 @@ class _RenderContext:
                     new_metadata = utils.dataframe_fingerprint(value)
                 context.state_metadata[key] = new_metadata
                 # TODO: enable
-                # context.needs_render = True
+                context.needs_render = True
                 if self._rerender_needed is False:
+                    if DEBUG:
+                        trigger_stack = traceback.format_stack()
+
+                        self._rerender_needed_reasons.append(
+                            RerenderReason(
+                                reason=f"state changed with key {key}",
+                                prev_value=prev_value,
+                                next_value=value,
+                                created_stack=created_stack,
+                                trigger_stack=trigger_stack,
+                            )
+                        )
+                    else:
+                        self._rerender_needed_reasons.append(RerenderReason(reason=f"state changed with key {key}", prev_value=prev_value, next_value=value))
                     self._rerender_needed = True
-                    self._rerender_needed_reason = f"{key} changed"
                 if not self._is_rendering:
                     self.render(self.element, self.container)
                 else:
@@ -1175,7 +1275,6 @@ class _RenderContext:
             # we always set it, even replacing it when we didn't execute it
             # in the consolidation phase we decide what to do (e.g. skip it)
             logger.info("Setting next effect = %r for index %r (%r)", effect, self.context.effect_index, dependencies)
-            assert not previous_effect._cleaned_up
             if previous_effect.executed:
                 # line up...
                 previous_effect.next = Effect(effect, dependencies)
@@ -1187,8 +1286,8 @@ class _RenderContext:
     def update(self, element: Element):
         if self._is_rendering:
             self.element = element
+            self._rerender_needed_reasons.append(RerenderReason(reason="root element changed"))
             self._rerender_needed = True
-            self._rerender_needed_reason = "root element changed"
         else:
             self.render(element, self.container)
 
@@ -1217,7 +1316,6 @@ class _RenderContext:
                 main_render_phase = not self._is_rendering
                 render_count = self.render_count  # make a copy
                 self._rerender_needed = False
-                self._rerender_needed_reason = None
                 logger.info("Render phase: %r %r of %r", self.render_count, "main" if main_render_phase else "(nested)", self.element)
                 self.render_count += 1
                 self._is_rendering = True
@@ -1246,17 +1344,29 @@ class _RenderContext:
                         # but if an exception bubbled up, we should stop
                         while self._rerender_needed and not self.context_root.exceptions_children:
                             if render_counts > 50:
-                                raise RuntimeError(f"Too many renders triggered, your render loop does not stop, last reason: {self._rerender_needed_reason}")
-                            logger.info("Entering nested render phase: %r", self._rerender_needed_reason)
+
+                                def format(reason: RerenderReason):
+                                    f = f"Reason: {reason.reason}\nValue changed from {reason.prev_value} to {reason.next_value}\n"
+                                    if reason.created_stack:
+                                        f += f"Created at: {''.join(reason.created_stack)}\n"
+                                    if reason.trigger_stack:
+                                        f += f"Triggered at: {''.join(reason.trigger_stack)}\n"
+                                    return f
+
+                                self._rerender_needed_reasons[-1]
+                                msg = f"Too many renders triggered, your render loop does not stop\nLast reason: {format(self._rerender_needed_reasons[-1])}\n"
+                                if len(self._rerender_needed_reasons) >= 2:
+                                    msg += f"Previous reasons: {format(self._rerender_needed_reasons[-2])}\n"
+                                raise RuntimeError(msg)
+                            logger.info("Entering nested render phase: %r", self._rerender_needed_reasons[-1])
                             self._rerender_needed = False
-                            self._rerender_needed_reason = None
                             self._shared_elements_next = set()
                             self.context.exception_handler = False
                             self.context.exceptions_children = []
                             self.context.exceptions_self = []
 
                             self._render(self.element, "/", parent_key=ROOT_KEY)
-                            logger.info("Render done: %r %r", self._rerender_needed, self._rerender_needed_reason)
+                            logger.info("Render done: %r %r", self._rerender_needed, self._rerender_needed_reasons[-1])
                             assert self.context is self.context_root
                             render_counts += 1
                         logger.debug("Render phase resulted in (next) elements:")
@@ -1308,7 +1418,7 @@ class _RenderContext:
                             break
 
                         if self._rerender_needed:
-                            logger.info("Need rerender after reconsolidation: %r", self._rerender_needed_reason)
+                            logger.info("Need rerender after reconsolidation: %r", self._rerender_needed_reasons[-1])
                             stable = False
                         else:
                             stable = True
@@ -1348,7 +1458,7 @@ class _RenderContext:
                     value = html.escape(error)
                 from . import ipywidgets as w
 
-                return self.render(w.HTML(value="<pre>" + value + "</pre>"), self.container)
+                return self.render(w.HTML(value="<pre>" + value + "</pre>", layout=w.Layout(overflow="auto")), self.container)
             else:
                 raise exc
         return widget
@@ -1364,7 +1474,7 @@ class _RenderContext:
         if default_key == "/":
             # if this is the root element, reset
             context.used_keys.clear()
-            default_key = element.component.name + "/"
+            default_key = "/"
 
         el = element
         # if we did not define a custom key, use the default key
@@ -1387,12 +1497,18 @@ class _RenderContext:
                 return
             else:
                 self._shared_elements_next.add(el)
+        el_prev = context.elements_next.get(key)
+        if el_prev is None:
+            el_prev = context.elements.get(key)
         context.elements_next[key] = el
         # used for testing
         el._render_count += 1
 
         if isinstance(el.component, ComponentWidget):
             assert not el.args, "no positional args supported for widgets"
+            # if at this place we had a componentfunction, remove it
+            if key in context.children_next:
+                del context.children_next[key]
 
         if el.args or el.kwargs:
             # do this conditionally to make logs cleaner
@@ -1425,6 +1541,7 @@ class _RenderContext:
                         logger.debug("Render: Not the same component, we just copy the children and elements of the ComponentContext")
                         # The old context is cleaned up in the reconciliation phase
                         context = ComponentContext(parent=parent_context, context_managers=[cm(el) for cm in _component_context_manager_classes])
+                        el_prev = None  # we dont want to compare the old element
                     else:
                         logger.debug("Render: Same component: %r", el.component)
                         context = context_previous
@@ -1437,66 +1554,85 @@ class _RenderContext:
             else:
                 logger.debug("Render: New ComponentContext")
                 context = ComponentContext(parent=parent_context, context_managers=[cm(el) for cm in _component_context_manager_classes])
+                el_prev = None  # we dont want to compare the old element
             context.invoke_element = el
             assert context.parent is not None
             self.container_adders = []
             logger.debug("Render: Enter context %r and excuting component function %r", key, el.component.f)
             self.context = context
             render_count = self.render_count
+            needs_render = context.needs_render
+            if not needs_render:
+                if el_prev is not None and context_previous is context:
+                    assert not isinstance(el_prev.component, ComponentWidget)
+                    needs_render = el._arguments_changed(el_prev)
+                if context.exceptions_children:
+                    # we have exceptions, so we need to render
+                    needs_render = True
+            if not needs_render:
+                assert el_prev is not None
             try:
-                context.state_index = 0
-                context.effect_index = 0
-                context.memo_index = 0
-                context.user_contexts = {}
-                context.exception_handler = False
                 # this is reset in use_exception
                 # context.exceptions_children = []
                 # TODO: why do the tests pass if we comment the next line out
                 context.exceptions_self = []
-                # when we have nested renders, we can already have this filled
-                context.elements_next.clear()
-                # Now, we actually execute the render function, and get
-                # back the root element
-                root_element: Optional[Element] = None
-                try:
-                    stack = contextlib.ExitStack()
-                    with contextlib.ExitStack() as stack:
-                        for cm in context.context_managers:
-                            stack.enter_context(cm)
-                        if _default_container is not None:
-                            with _default_container() as container:
+                if needs_render:
+                    context.state_index = 0
+                    context.effect_index = 0
+                    context.memo_index = 0
+                    context.user_contexts = {}
+                    context.exception_handler = False
+                    # we reset if before calling the component
+                    # which might set it to true again
+                    context.needs_render = False
+                    # Now, we actually execute the render function, and get
+                    # back the root element
+                    root_element: Optional[Element] = None
+                    try:
+                        stack = contextlib.ExitStack()
+                        with contextlib.ExitStack() as stack:
+                            for cm in context.context_managers:
+                                stack.enter_context(cm)
+                            if _default_container is not None:
+                                with _default_container() as container:
+                                    el.component.render_count += 1
+                                    root_element = el.component.f(*el.args, **el.kwargs)
+                                if root_element is None:
+                                    if len(container.kwargs["children"]) == 1:
+                                        root_element = container.kwargs["children"][0]
+                                    else:
+                                        root_element = container
+                            else:
+                                el.component.render_count += 1
                                 root_element = el.component.f(*el.args, **el.kwargs)
-                            if root_element is None:
-                                if len(container.kwargs["children"]) == 1:
-                                    root_element = container.kwargs["children"][0]
-                                else:
-                                    root_element = container
-                        else:
-                            root_element = el.component.f(*el.args, **el.kwargs)
-                except BaseException as e:
-                    if DEBUG:
-                        # we might be interested in the traceback inside the call...
-                        if len(self.tracebacks) == 0:
-                            assert e.__traceback__ is not None
-                            traceback = cast(TracebackType, e.__traceback__)
-                            if traceback.tb_next:  # is there an error inside the call
-                                self.tracebacks.append(traceback.tb_next)
-                        self.tracebacks.append(el.traceback)
-                    logger.exception("Component %r raised exception %r", el.component, e)
-                    context.exceptions_self.append(e)
-                    self._rerender_needed_reason = "Exception ocurred during render"
-                    self._rerender_needed = True
+                            assert root_element is not None
+                    except BaseException as e:
+                        if DEBUG:
+                            # we might be interested in the traceback inside the call...
+                            if len(self.tracebacks) == 0:
+                                assert e.__traceback__ is not None
+                                traceback = cast(TracebackType, e.__traceback__)
+                                if traceback.tb_next:  # is there an error inside the call
+                                    self.tracebacks.append(traceback.tb_next)
+                            self.tracebacks.append(el.traceback)
+                        logger.exception("Component %r raised exception %r", el.component, e)
+                        context.exceptions_self.append(e)
+                        self._rerender_needed_reasons.append(RerenderReason(reason="Exception ocurred during render"))
+                        self._rerender_needed = True
+                        context.needs_render = True
 
-                if root_element is None and not context.exceptions_self:
-                    raise ValueError(f"Component {el.component} returned None")
+                    if root_element is None and not context.exceptions_self:
+                        raise ValueError(f"Component {el.component} returned None")
+                else:
+                    root_element = context.root_element_next or context.root_element
 
                 if self.render_count != render_count:
                     raise RuntimeError("Recursive render detected, possible a bug in react")
                 if root_element is not None:
                     logger.debug("root element: %r %x", root_element, id(root_element))
-                    context.root_element_next = root_element
                     new_parent_key = join_key(parent_key, key)
                     self._render(root_element, "/", parent_key=new_parent_key)  # depth first
+                    context.root_element_next = root_element
                 else:
                     if el.is_shared:
                         # TODO: why do the tests pass if we comment the next line out
@@ -1518,12 +1654,16 @@ class _RenderContext:
                 except RuntimeError as e:
                     logger.exception("Exception in hook count check")
                     context.exceptions_self.append(e)
-                    self._rerender_needed_reason = "Exception ocurred during render (hook count check)"
+                    self._rerender_needed_reasons.append(RerenderReason(reason="Exception ocurred during render (hook count check)"))
                     self._rerender_needed = True
+                    context.needs_render = True
                 # only expose to parent when no error occurs
                 context.parent.children_next[key] = context
                 # drop all children from the previous render run (this render phase)
                 context.children_next = {k: v for k, v in context.children_next.items() if k in context.used_keys}
+                # same for elements
+                context.elements_next = {k: v for k, v in context.elements_next.items() if k in context.used_keys}
+                context.user_contexts_prev = context.user_contexts
             finally:
                 assert context.parent is parent_context
                 self.context = context.parent
@@ -1536,8 +1676,9 @@ class _RenderContext:
                         # this happens when an exception was added from an event handler
                         # this means no exception was raised during the render phase
                         # but we still need to rerender, until someone catches the exception
-                        self._rerender_needed_reason = "Exception ocurred during during event handler"
+                        self._rerender_needed_reasons.append(RerenderReason(reason="Exception ocurred during render"))
                         self._rerender_needed = True
+                        self.context.needs_render = True
 
             assert context is not None
 
@@ -1547,7 +1688,7 @@ class _RenderContext:
         key = el._key
         if key is None:
             if default_key == "/":
-                default_key = el.component.name + "/"
+                default_key = "/"
             key = default_key
         assert key is not None
         logger.debug("Reconsolidate: (%s,%s) %r", parent_key, key, el)
@@ -1627,31 +1768,35 @@ class _RenderContext:
                             # if we have a next, it means that effect itself is executed
                             # TODO: custom equals
                             if effect.next.dependencies is not None and utils.equals(effect.dependencies, effect.next.dependencies):
-                                logger.info("No need to add effect, dependencies are the same (%r)", effect.dependencies)
+                                logger.info("No need to add effect, dependencies are the same (%r %r)", effect.callable, effect.dependencies)
                                 # not needed, just remove the reference
                                 effect.next = None
                             else:
                                 # dependencies changed, cleanup and execute next
-                                try:
-                                    effect.cleanup()
-                                except BaseException as e:
-                                    context.exceptions_self.append(e)
-                                    self._rerender_needed_reason = "Exception ocurred during effect"
-                                    self._rerender_needed = True
+                                if not effect._cleaned_up:
+                                    try:
+                                        effect.cleanup()
+                                    except BaseException as e:
+                                        context.exceptions_self.append(e)
+                                        self._rerender_needed_reasons.append(RerenderReason(reason="Exception ocurred during effect"))
+                                        self._rerender_needed = True
+                                        context.needs_render = True
                                 effect = child_context.effects[effect_index] = effect.next
                                 try:
                                     effect()
                                 except BaseException as e:
                                     context.exceptions_self.append(e)
-                                    self._rerender_needed_reason = "Exception ocurred during effect"
+                                    self._rerender_needed_reasons.append(RerenderReason(reason="Exception ocurred during effect"))
                                     self._rerender_needed = True
+                                    context.needs_render = True
                         else:
                             try:
                                 effect()
                             except BaseException as e:
                                 context.exceptions_self.append(e)
-                                self._rerender_needed_reason = "Exception ocurred during effect"
+                                self._rerender_needed_reasons.append(RerenderReason(reason="Exception ocurred during effect"))
                                 self._rerender_needed = True
+                                context.needs_render = True
 
                     if child_context.children_next:
                         # if we had two render phases, we can have old context left over
@@ -1723,7 +1868,7 @@ class _RenderContext:
                         el._update_widget(widget_previous, el_prev, kwargs)
                     except BaseException as e:
                         context.exceptions_self.append(e)
-                        self._rerender_needed_reason = "Exception ocurred during reconciliation (updating widget)"
+                        self._rerender_needed_reasons.append(RerenderReason(reason="Exception ocurred during reconciliation (updating widget)"))
                         self._rerender_needed = True
                     if el.is_shared:
                         self._shared_widgets[el] = widget_previous
@@ -1816,7 +1961,7 @@ class _RenderContext:
         key = el._key
         if key is None:
             if default_key == "/":
-                default_key = el.component.name + "/"
+                default_key = "/"
             key = default_key
         assert key is not None
         assert self.context is not None
@@ -1846,10 +1991,11 @@ class _RenderContext:
 
                 for effect_index, effect in enumerate(self.context.effects):
                     try:
-                        effect.cleanup()
+                        if not effect._cleaned_up:
+                            effect.cleanup()
                     except BaseException as e:
                         child_context.exceptions_self.append(e)
-                        self._rerender_needed_reason = "Exception ocurred during effect"
+                        self._rerender_needed_reasons.append(RerenderReason(reason="Exception ocurred during effect"))
                         self._rerender_needed = True
                 assert self.context.root_element is not None
                 new_parent_key = join_key(parent_key, key)
@@ -1873,10 +2019,10 @@ class _RenderContext:
 
             def close(widget: widgets.Widget):
                 # this happens for v.Chip, which has a close trait
-                if not callable(widget.close):
-                    super(widget.__class__, widget).close()  # type: ignore
-                else:
+                if callable(widget.close):
                     widget.close()
+                else:
+                    logger.warning("Widget %r does not have a close method, possibly a close trait was added", widget)
 
             for orphan in self._orphans.get(widget.model_id, set()):
                 orphan_widget = widgets.Widget.widgets.get(orphan)
