@@ -1,3 +1,4 @@
+import sys
 import time
 import traceback
 import unittest.mock
@@ -19,6 +20,7 @@ from reacton.core import component, use_effect
 
 from . import bqplot  # noqa: F401
 from . import logging  # noqa: F401
+from . import core
 from . import ipyvuetify as v
 from . import ipywidgets as w
 from .core import ipywidget_version_major
@@ -41,6 +43,28 @@ def count():
     return len(widgets.Widget.widgets)
 
 
+class ContextManager:
+    def __init__(self, el) -> None:
+        self.el = el
+
+    def __enter__(self):
+        pass
+
+    def __exit__(self, *args):
+        def effect():
+            def cleanup():
+                pass
+
+            return cleanup
+
+        reacton.use_effect(effect, [])
+
+
+def set_text(text: str):
+    pass
+
+
+reacton.core._component_context_manager_classes.append(ContextManager)
 # components used for testing
 
 
@@ -118,9 +142,9 @@ def test_internals():
     assert rc.context_root.root_element == app
     assert rc.context_root.root_element_next is None
     assert list(rc.context_root.children_next) == []
-    assert list(rc.context_root.children) == ["App/"]
+    assert list(rc.context_root.children) == ["/"]
 
-    app_context = rc.context_root.children["App/"]
+    app_context = rc.context_root.children["/"]
     assert list(app_context.children_next) == []
     assert list(app_context.children) == ["child"]
     assert app_context.invoke_element is app
@@ -351,7 +375,33 @@ def test_render_replace():
     rc.close()
 
 
-def test_render_replace_component_with_element():
+def test_render_conditional_child():
+
+    set_value = lambda x: None  # noqa
+
+    @react.component
+    def Child():
+        nonlocal set_value
+        value, set_value = react.use_state(0)
+        if value == 0:
+            return w.Button(description="Button in child")
+        else:
+            return w.Label(description="Label in child")
+
+    @react.component
+    def Test():
+        return Child()
+
+    hbox, rc = react.render(Test(), handle_error=False)
+    assert rc._find(widgets.Button).widget.description == "Button in child"
+    set_value(1)
+    assert rc._find(widgets.Label).widget.description == "Label in child"
+    set_value(0)
+    assert rc._find(widgets.Button).widget.description == "Button in child"
+    rc.close()
+
+
+def test_render_conditional_replace_component_with_element():
 
     set_value = lambda x: None  # noqa
 
@@ -822,7 +872,7 @@ def test_use_effect_no_deps():
     cleanups = 0
 
     @react.component
-    def TestNoDeps(a, b):
+    def TestNoDeps(a, b, not_used=0):
         def test_effect():
             def cleanup():
                 nonlocal cleanups
@@ -838,7 +888,7 @@ def test_use_effect_no_deps():
     hbox, rc = react.render(TestNoDeps(a=1, b=1))
     assert calls == 1
     assert cleanups == 0
-    rc.render(TestNoDeps(a=1, b=1), hbox)
+    rc.render(TestNoDeps(a=1, b=1, not_used=2), hbox)
     assert calls == 2
     assert cleanups == 1
     rc.close()
@@ -1132,7 +1182,7 @@ def test_context():
         return w.Button(description=f"Child1: Clicked {clicks} times", on_click=lambda: dispatch("increment"))
 
     @react.component
-    def App():
+    def App(external_state=0):
         clicks, dispatch = react.use_reducer(click_reducer, 0)
         assert react.use_context(store_context) is None
         store_context.provide((clicks, dispatch))
@@ -1141,20 +1191,32 @@ def test_context():
             Child2()
         return main
 
-    hbox, rc = react.render_fixed(App())
-    button1 = hbox.children[0]
-    button2 = hbox.children[1]
+    hbox, rc = react.render(App(), handle_error=False)
+    button1 = hbox.children[0].children[0]
+    button2 = hbox.children[0].children[1]
     assert button1.description == "Child1: Clicked 0 times"
     assert button2.description == "Child2: Clicked 0 times"
+    assert Child1.render_count == 1  # type: ignore
+    assert SubChild2.render_count == 1  # type: ignore
     button1.click()
     assert button1.description == "Child1: Clicked 1 times"
     assert button2.description == "Child2: Clicked 1 times"
+    assert Child1.render_count == 2  # type: ignore
+    assert SubChild2.render_count == 2  # type: ignore
     button2.click()  # not attached
     assert button1.description == "Child1: Clicked 1 times"
     assert button2.description == "Child2: Clicked 1 times"
     button1.click()
     assert button1.description == "Child1: Clicked 2 times"
     assert button2.description == "Child2: Clicked 2 times"
+    assert Child1.render_count == 3  # type: ignore
+    assert SubChild2.render_count == 3  # type: ignore
+
+    # shouldn't re-render the children
+    rc.render(App(external_state=1))
+    assert Child1.render_count == 3  # type: ignore
+    assert SubChild2.render_count == 3  # type: ignore
+
     rc.close()
 
 
@@ -1848,7 +1910,11 @@ def test_recover_exception_in_cleanup_child(Container1, Container2):
     assert rc.find(ipywidgets.IntSlider)
     with pytest.raises(Exception):
         rc.render(Container1(children=[Container2()]))
-    rc.close()
+
+    try:
+        rc.close()
+    except Exception as e:
+        assert e.args[0] == "fail"
 
 
 def test_recover_exception_in_cleanup():
@@ -1923,9 +1989,9 @@ def test_state_get():
     slider, rc = react.render_fixed(Test())
     assert set_value is not None
     state = rc.state_get()
-    assert state == {"children": {"Test/": {"state": {"0": 0}}}, "state": {}}
+    assert state == {"children": {"/": {"state": {"0": 0}}}, "state": {}}
     set_value(42)
-    assert state == {"children": {"Test/": {"state": {"0": 42}}}, "state": {}}
+    assert state == {"children": {"/": {"state": {"0": 42}}}, "state": {}}
     assert slider.value == 42
     rc.close()
 
@@ -2839,3 +2905,97 @@ def test_component_context_manager():
 def test_close_when_overridden():
     box, rc = react.render(v.Chip(), handle_error=False)
     rc.close()
+    box, rc = react.render(v.Chip(), handle_error=False)
+    rc.close()
+
+
+def test_render_perf_child_only():
+
+    render_child = unittest.mock.Mock()
+    render_main = unittest.mock.Mock()
+    set_text = lambda x: None  # noqa
+
+    @react.component
+    def ClickButton():
+        render_child()
+        nonlocal set_text
+        text, set_text = react.use_state("initial")
+        return w.Button(description=text)
+
+    @react.component
+    def Test():
+        render_main()
+        return ClickButton()
+
+    box, rc = react.render(Test(), handle_error=False)
+    assert len(rc._find(widgets.Button)) == 1
+    assert render_child.call_count == 1
+    assert render_main.call_count == 1
+    assert rc._find(widgets.Button).widget.description == "initial"
+    set_text("new")
+    assert render_child.call_count == 2
+    assert render_main.call_count == 1
+    assert rc._find(widgets.Button).widget.description == "new"
+    rc.close()
+
+
+def test_render_repeated(ButtonComponent):
+    set_state = lambda x: None  # noqa
+
+    @reacton.component
+    def Test():
+        nonlocal set_state
+        state, set_state = reacton.use_state(0)
+        if state == 0:
+            return w.Button(description="hoeba")
+        if state == 1:
+            with w.VBox() as main:
+                w.Button(description="hoeba1")
+                w.Button(description="hoeba2")
+                ButtonComponent()
+            set_state(0)
+            return main
+
+    box, rc = react.render(Test(), handle_error=False)
+    assert len(rc.find(widgets.Button, description="hoeba")) == 1
+    set_state(1)
+    assert len(rc.find(widgets.Button, description="hoeba")) == 1
+    rc.close()
+
+
+# does not run on Python 3.6
+@pytest.mark.skipif(sys.version_info < (3, 7), reason="requires python3.7 or higher")
+def test_debug_set_state():
+    set_state = None
+    core.DEBUG = True
+
+    @reacton.component
+    def Test():
+        nonlocal set_state
+        state, set_state = reacton.use_state(0)  # i should be in the reason stacktrace
+        return w.Button(description=str(state))
+
+    box, rc = react.render(Test(), handle_error=False)
+    assert rc.find(widgets.Button).widget.description == "0"
+    assert set_state is not None
+    set_state(2)
+    assert any(["i should be in the reason stacktrace" in k for k in rc._rerender_needed_reasons[-1].created_stack])
+    core.DEBUG = False
+    rc.close()
+
+
+def test_debug_infinite_loop():
+    @reacton.component
+    def Infinite():
+        state, set_state = reacton.use_state(0)
+        set_state(state + 1)
+        return w.Button(description=str(state))
+
+    with pytest.raises(RuntimeError) as e:
+        rc = core._RenderContext(Infinite(), handle_error=False)
+        rc.render(rc.element)
+
+    assert "Too many renders triggered" in str(e)
+    assert len(rc._rerender_needed_reasons) >= 50
+
+    rc.close()
